from __future__ import annotations
from typing import Optional, Callable, List

import torch
from torch.distributions import Distribution
from abc import ABC, abstractmethod
import storch
from storch import Plate


class SamplingMethod(ABC, torch.nn.Module):
    def __init__(self, plate_name: str):
        super().__init__()
        self.reset()
        self.plate_name = plate_name

    def reset(self) -> None:
        pass

    def forward(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        requires_grad: bool,
    ) -> (storch.StochasticTensor, Plate):
        return self.sample(distr, parents, plates, requires_grad)

    @abstractmethod
    def sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        requires_grad: bool,
    ) -> (storch.StochasticTensor, Plate):
        pass

    def weighting_function(
        self, tensor: storch.StochasticTensor, plate: Plate
    ) -> Optional[storch.Tensor]:
        """
        Weight by the size of the sample. Overload this if your gradient estimation uses some kind of weighting
        of the different events, like importance sampling or computing the expectation.
        If None is returned, it is assumed the samples are iid monte carlo samples.
        """
        return self.mc_weighting_function(tensor, plate)

    def mc_weighting_function(
        self, tensor: storch.StochasticTensor, plate: Plate
    ) -> Optional[storch.Tensor]:
        return None

    def update_parameters(
        self,
        result_triples: [(storch.StochasticTensor, storch.CostTensor, torch.Tensor)],
    ):
        pass

    def mc_sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ) -> torch.Tensor:
        return distr.sample((self.n_samples,))

    def set_mc_sample(
        self,
        new_sample_func: Callable[
            [Distribution, [storch.Tensor], [Plate], int], torch.Tensor
        ],
    ) -> SamplingMethod:
        """
        Override storch.Method specific sampling functions.
        This is called when initializing a storch.Method that has slightly different MC sampling semantics
        (for example, reparameterization instead of normal sampling).
        This allows for compatibility of different `storch.Method`'s with different `storch.sampling.Method`'s.
        """
        self.mc_sample = new_sample_func
        return self

    def set_mc_weighting_function(
        self,
        new_weighting_func: Callable[
            [storch.StochasticTensor, Plate], Optional[storch.Tensor]
        ],
    ) -> SamplingMethod:
        """
        Override storch.Method specific weighting functions.
        This is called when initializing a storch.Method that has slightly different MC weighting semantics
        (for example, REBAR that weights some samples differently).
        This allows for compatibility of different `storch.Method`'s with different `storch.sampling.Method`'s.
        """
        self.mc_weighting_function = new_weighting_func
        return self

    def on_plate_already_present(self, plate: Plate):
        raise ValueError(
            "Cannot create stochastic tensor with name "
            + plate.name
            + ". A parent sample has already used this name. Use a different name for this sample."
        )


class MonteCarlo(SamplingMethod):
    """
    Monte Carlo sampling methods use simple sampling methods that take n independent samples.
    Unlike complex ancestral sampling methods such as SampleWithoutReplacementMethod, the sampling behaviour is not dependent
    on earlier samples in the stochastic computation graph (but the distributions are!).
    """

    def __init__(self, plate_name: str, n_samples: int = 1):
        super().__init__(plate_name)
        self.n_samples = n_samples

    def sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        requires_grad: bool,
    ) -> (storch.StochasticTensor, Plate):
        plate = None
        for _plate in plates:
            if _plate.name == self.plate_name:
                plate = _plate
                break
        n_samples = 1 if plate else self.n_samples
        with storch.ignore_wrapping():
            tensor = self.mc_sample(distr, parents, plates, n_samples)
        plate_size = tensor.shape[0]
        if tensor.shape[0] == 1:
            tensor = tensor.squeeze(0)

        if not plate:
            plate = Plate(self.plate_name, plate_size, plates.copy())
            plates.insert(0, plate)

        if isinstance(tensor, storch.Tensor):
            tensor = tensor._tensor

        s_tensor = storch.StochasticTensor(
            tensor, parents, plates, self.plate_name, plate_size, distr, requires_grad,
        )
        return s_tensor, plate



from typing import Optional, Tuple

from torch.distributions import (
    Distribution,
    Bernoulli,
    TransformedDistribution,
    Uniform,
    SigmoidTransform,
    AffineTransform,
)
import torch
import storch
from storch import Plate, StochasticTensor, CostTensor
from storch.method import Method
from storch.sampling import SamplingMethod, MonteCarlo
from storch.util import magic_box


class Logistic(TransformedDistribution):
    def __init__(self, loc: float, scale: float):
        # Copied from https://pytorch.org/docs/stable/distributions.html#torch.distributions.transformed_distribution.TransformedDistribution
        super().__init__(
            Uniform(0, 1),
            transforms=[SigmoidTransform().inv, AffineTransform(loc=loc, scale=scale)],
        )


class ARM(Method):
    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(
            plate_name, sampling_method.set_mc_sample(self.sample_arm).set_mc_weighting_function(self.weighting_function),
        )

    def sample_arm(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ):
        if not isinstance(distr, Bernoulli):
            raise ValueError("ARM only works for the Bernoulli distribution.")
        logits = distr.logits
        eps = Logistic(0.0, 1.0).sample((amt_samples,) + logits.shape)
        z = eps + logits
        z_tilde = -eps + logits
        return torch.cat([z, z_tilde], dim=0)

    def weighting_function(
        self, tensor: storch.StochasticTensor, plate: Plate
    ) -> Optional[storch.Tensor]:
        # The antithetic sample is not according to the true distribution, so we cannot count it during the weighting
        n = int(tensor.n / 2)
        weighting = tensor._tensor.new_zeros((tensor.n,))
        weighting[:n] = tensor._tensor.new_tensor(1.0 / n)
        return weighting

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        return tensor > 0.0

    def comp_estimator(
        self, tensor: torch.Tensor, cost: torch.Tensor, logits: torch.Tensor, plate_index: int, n: int
    ) -> Tuple[storch.Tensor, storch.Tensor]:
        _index_z = (slice(None),) * plate_index + (slice(n),)
        _index_z_tilde = (slice(None),) * plate_index + (slice(n, 2 * n),)
        z = tensor[_index_z]

        baseline = torch.zeros_like(cost)
        baseline[_index_z] = 0.5 * (cost[_index_z] + cost[_index_z_tilde])

        logistic = Logistic(logits, 1.0)
        log_prob = torch.zeros_like(tensor)
        log_prob[_index_z] = logistic.log_prob(z.detach())
        return log_prob, baseline

    def estimator(
        self, tensor: StochasticTensor, cost: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        # TODO: No support for alternative plate weighting
        plate = tensor.get_plate(tensor.name)
        index = tensor.get_plate_dim_index(plate.name)
        log_prob, baseline = storch.deterministic(self.comp_estimator)(
            tensor, cost, tensor.distribution.logits, index, plate.n // 2
        )

        return log_prob, (1-magic_box(log_prob)) * baseline


class DisARM(Method):
    """
    Introduces by Dong, Mnih and Tucker, 2020 https://arxiv.org/abs/2006.10680
    """

    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(
            plate_name, sampling_method.set_mc_sample(self.sample_disarm),
        )

    def sample_disarm(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ):
        if not isinstance(distr, Bernoulli):
            raise ValueError("DisARM only works for the Bernoulli distribution.")
        logits = distr.logits
        eps = Logistic(0.0, 1.0).sample((amt_samples,) + logits.shape)
        z = eps + logits
        z_tilde = -eps + logits
        # Sample (b, \tilde{b})
        return torch.cat([z, z_tilde], dim=0) > 0.0

    def plate_weighting(
        self, tensor: storch.StochasticTensor, plate: Plate
    ) -> Optional[storch.Tensor]:
        # The antithetic sample is not according to the true distribution, so we cannot count it during the weighting
        n = int(tensor.n / 2)
        weighting = tensor._tensor.new_zeros((tensor.n,))
        weighting[:n] = tensor._tensor.new_tensor(1.0 / n)
        return weighting

    @storch.deterministic
    def comp_estimator(
        self, tensor: StochasticTensor, cost: CostTensor, logits, plate_index, n
    ) -> Tuple[storch.Tensor, storch.Tensor]:
        _index_z = (None,) * plate_index + (slice(n),)
        _index_ztilde = (None,) * plate_index + (slice(n, 2 * n),)
        f_z_tilde = cost[_index_ztilde]
        b = tensor[_index_z]
        b_tilde = tensor[_index_ztilde]

        weighting = ((-1) ** b_tilde.float()) * ~b.eq(b_tilde) * logits.abs().sigmoid()
        log_prob = torch.zeros_like(tensor, dtype=cost.dtype)
        log_prob[_index_z] = weighting.detach() * logits

        baseline = torch.zeros_like(cost)
        baseline[_index_z] = f_z_tilde

        return log_prob, baseline

    def estimator(
        self, tensor: StochasticTensor, cost: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        # TODO: No support for alternative plate weighting
        # TODO: This doesn't follow the current implementation
        plate = tensor.get_plate(tensor.name)
        index = tensor.get_plate_dim_index(plate.name)
        multiplicative, baseline = self.comp_estimator(
            tensor, cost, tensor.distribution.logits, index, plate.n // 2
        )

        return (
            0.5 * multiplicative.sum(dim=multiplicative.event_dim_indices),
            baseline,
        )
